/*
 * Copyright (C) 2006 Ruslan Popov <radz@yandex.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "hsv.h"

// Widget's handlers
static void on_bReset_clicked( GtkButton *button, gpointer p_this )
{
  ( ( HSVPlugin * ) p_this )->ResetValues();
  GetSelectedFramesForFX().Repaint();
}

static void on_hsHue_changed( GtkRange *range, gpointer p_this )
{
  ( ( HSVPlugin * ) p_this )->Scale2Spin( "hsHue", "sbHue" );
  GetSelectedFramesForFX().Repaint();
}

static void on_sbHue_changed( GtkRange *range, gpointer p_this )
{
  ( ( HSVPlugin * ) p_this )->Spin2Scale( "sbHue", "hsHue" );
  GetSelectedFramesForFX().Repaint();
}

static void on_hsSaturation_changed( GtkRange *range, gpointer p_this )
{
  ( ( HSVPlugin * ) p_this )->Scale2Spin( "hsSaturation", "sbSaturation" );
  GetSelectedFramesForFX().Repaint();
}

static void on_sbSaturation_changed( GtkRange *range, gpointer p_this )
{
  ( ( HSVPlugin * ) p_this )->Spin2Scale( "sbSaturation", "hsSaturation" );
  GetSelectedFramesForFX().Repaint();
}

static void on_hsValue_changed( GtkRange *range, gpointer p_this )
{
  ( ( HSVPlugin * ) p_this )->Scale2Spin( "hsValue", "sbValue" );
  GetSelectedFramesForFX().Repaint();
}

static void on_sbValue_changed( GtkRange *range, gpointer p_this )
{
  ( ( HSVPlugin * ) p_this )->Spin2Scale( "sbValue", "hsValue" );
  GetSelectedFramesForFX().Repaint();
}

// Constructor
HSVPlugin::HSVPlugin()
{
  // create a widget for plugin
  glade = glade_xml_new( DATADIR "/" PACKAGE "/hsv.glade", NULL, NULL );
  plugin_window = glade_xml_get_widget( glade, "hsv" );
  // connect widgets' handlers
  GtkWidget* widget;
  widget = glade_xml_get_widget( glade, "bReset" );
  if ( widget ) {
    g_signal_connect( G_OBJECT( widget ), "clicked",
                      G_CALLBACK( on_bReset_clicked ), this );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "hsHue" );
  if ( widget ) {
    g_signal_connect( G_OBJECT( widget ), "value-changed",
                      G_CALLBACK( on_hsHue_changed ), this );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "sbHue" );
  if ( widget ) {
    g_signal_connect( G_OBJECT( widget ), "value-changed",
                      G_CALLBACK( on_sbHue_changed ), this );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "hsSaturation" );
  if ( widget ) {
    g_signal_connect( G_OBJECT( widget ), "value-changed",
                      G_CALLBACK( on_hsSaturation_changed ), this );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "sbSaturation" );
  if ( widget ) {
    g_signal_connect( G_OBJECT( widget ), "value-changed",
                      G_CALLBACK( on_sbSaturation_changed ), this );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "hsValue" );
  if ( widget ) {
    g_signal_connect( G_OBJECT( widget ), "value-changed",
                      G_CALLBACK( on_hsValue_changed ), this );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "sbValue" );
  if ( widget ) {
    g_signal_connect( G_OBJECT( widget ), "value-changed",
                      G_CALLBACK( on_sbValue_changed ), this );
    widget = NULL;
  }
}

// Destructor
HSVPlugin::~HSVPlugin()
{
  // remove our widget
  gtk_widget_destroy( plugin_window );
}

char * HSVPlugin::GetDescription( ) const
{
  return "HSV Color Space";
}

/** Here we filter our frame. This function converts frame's pixels to
    HSV colorspace, applies user's changes to HSV values (we do not
    forget to check edge conditions as usual) and converts pixels
    back to RGB colorspace.
 */
void HSVPlugin::FilterFrame( uint8_t *pixels, int width, int height,
				    double position, double delta_frame )
{
  hsvcolor hsv;
  uint8_t * length = pixels + ( width * height * 3 );
  while ( length >= pixels )
    {
      rgb2hsv( length, hsv );
      hsv[H] += hue;
      hsv[S] += saturation;
      hsv[V] += value;
      if ( hsv[H] > 1 ) hsv[H] -= 1;
      if ( hsv[H] < 0 ) hsv[H] += 1;
      if ( hsv[S] > 1 ) hsv[S] = 1;
      if ( hsv[S] < 0 ) hsv[S] = 0;
      if ( hsv[V] > 1 ) hsv[V] = 1;
      if ( hsv[V] < 0 ) hsv[V] = 0;
      hsv2rgb( length, hsv );
      length-=3;
    }
}

void HSVPlugin::AttachWidgets( GtkBin *bin )
{
  gtk_widget_reparent( ( GTK_BIN( plugin_window ) )->child, GTK_WIDGET( bin ) );
}

void HSVPlugin::DetachWidgets( GtkBin *bin )
{
  gtk_widget_reparent( ( GTK_BIN( bin ) )->child, GTK_WIDGET( plugin_window ) );
}

/** This function is called after user is pressed on 'Preview' or
    'Render' button. This is useful place to make some initializing
    thing if it needs. */
void HSVPlugin::InterpretWidgets( GtkBin *bin )
{
  GtkWidget * spin;
  // get the hue from GUI control
  spin = glade_xml_get_widget( glade, "sbHue" );
  hue = ( spin ) ? gtk_spin_button_get_value( GTK_SPIN_BUTTON( spin ) ) / 100.0 : 0;
  spin = NULL;
  // get the saturation from GUI control
  spin = glade_xml_get_widget( glade, "sbSaturation" );
  saturation = ( spin ) ? gtk_spin_button_get_value( GTK_SPIN_BUTTON( spin ) ) / 100.0 : 0;
  spin = NULL;
  // get the value from GUI control
  spin = glade_xml_get_widget( glade, "sbValue" );
  value = ( spin ) ? gtk_spin_button_get_value( GTK_SPIN_BUTTON( spin ) ) / 100.0 : 0;
  spin = NULL;
}

/** This function is used to convert pixel to HSV colorspace. */
void HSVPlugin::rgb2hsv( uint8_t *pixel, hsvcolor hsv )
{
  double rgb[3];
  double cmin, cmax, delta;

  rgb[R] = pixel[R] / 255.0; 
  rgb[G] = pixel[G] / 255.0; 
  rgb[B] = pixel[B] / 255.0;

  cmin = minimum( minimum( rgb[R], rgb[G] ), rgb[B] );
  cmax = maximum( maximum( rgb[R], rgb[G] ), rgb[B] );
  delta = cmax - cmin;
 
  hsv[V] = cmax; // value

  if ( delta == 0.0 )
    {
      hsv[H] = hsv[S] = 0.0;
    }
  else
    {
      double dR, dG, dB;
      hsv[S] = delta / cmax; // saturation

      dR = ( ( ( cmax - rgb[R] ) / 6.0 ) + ( delta / 2.0 ) ) / delta;
      dG = ( ( ( cmax - rgb[G] ) / 6.0 ) + ( delta / 2.0 ) ) / delta;
      dB = ( ( ( cmax - rgb[B] ) / 6.0 ) + ( delta / 2.0 ) ) / delta;

      //hue
      if      ( rgb[R] == cmax ) hsv[H] = dB - dG;
      else if ( rgb[G] == cmax ) hsv[H] = ( 1.0 / 3.0 ) + dR - dB;
      else if ( rgb[B] == cmax ) hsv[H] = ( 2.0 / 3.0 ) + dG - dR;
      
      if ( hsv[H] < 0 ) hsv[H] += 1;
      if ( hsv[H] > 1 ) hsv[H] -= 1;
    }
}

/** This function is used to convert pixel to RGB colorspace. */
void HSVPlugin::hsv2rgb( uint8_t *pixel, hsvcolor hsv )
{
  int i;
  double rgb[3];
  double v1, v2, v3;
  double h, s, v;

  if ( hsv[S] == 0.0 )
    {
      // achromatic color: there is no hue
      rgb[R] = rgb[G] = rgb[B] = hsv[V];
    }
  else
    {
      h = hsv[H] * 6.0;
      if ( h == 6.0 ) h = 0.0;

      i = (int) ( h );

      /* find R, G and B components */
      v1 = hsv[V] * ( 1.0 - hsv[S] );
      v2 = hsv[V] * ( 1.0 - hsv[S] * ( h - i ) );
      v3 = hsv[V] * ( 1.0 - hsv[S] * ( 1.0 - ( h - i ) ) );

      switch ( i ) 
	{
	case 0: rgb[R] = hsv[V]; rgb[G] = v3;     rgb[B] = v1;     break;
	case 1: rgb[R] = v2;     rgb[G] = hsv[V]; rgb[B] = v1;     break;
	case 2: rgb[R] = v1;     rgb[G] = hsv[V]; rgb[B] = v3;     break;
	case 3: rgb[R] = v1;     rgb[G] = v2;     rgb[B] = hsv[V]; break;
	case 4: rgb[R] = v3;     rgb[G] = v1;     rgb[B] = hsv[V]; break;
	case 5: rgb[R] = hsv[V]; rgb[G] = v1;     rgb[B] = v2;     break;
	}
    }

  pixel[R] = (uint8_t) ( rgb[R] * 255 );
  pixel[G] = (uint8_t) ( rgb[G] * 255 );
  pixel[B] = (uint8_t) ( rgb[B] * 255 );
}

void HSVPlugin::ResetValues( void ) const
{
  GtkWidget * widget = NULL;
  // hue
  widget = glade_xml_get_widget( glade, "hsHue" );
  if ( widget ) {
    gtk_adjustment_set_value( gtk_range_get_adjustment( GTK_RANGE( widget ) ), 0.0 );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "sbHue" );
  if ( widget ) {
    gtk_spin_button_set_value( GTK_SPIN_BUTTON( widget ), 0.0 );
    widget = NULL;
  }
  // saturation
  widget = glade_xml_get_widget( glade, "hsSaturation" );
  if ( widget ) {
    gtk_adjustment_set_value( gtk_range_get_adjustment( GTK_RANGE( widget ) ), 0.0 );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "sbSaturation" );
  if ( widget ) {
    gtk_spin_button_set_value( GTK_SPIN_BUTTON( widget ), 0.0 );
    widget = NULL;
  }
  // value
  widget = glade_xml_get_widget( glade, "hsValue" );
  if ( widget ) {
    gtk_adjustment_set_value( gtk_range_get_adjustment( GTK_RANGE( widget ) ), 0.0 );
    widget = NULL;
  }
  widget = glade_xml_get_widget( glade, "sbValue" );
  if ( widget ) {
    gtk_spin_button_set_value( GTK_SPIN_BUTTON( widget ), 0.0 );
    widget = NULL;
  }
}

void HSVPlugin::Spin2Scale( const gchar * spinname, const gchar * scalename ) const
{
  // get spin value and set it to scale
  GtkWidget * scale = glade_xml_get_widget( glade, scalename );
  GtkWidget * spin = glade_xml_get_widget( glade, spinname );
  if ( scale && spin ) {
    gtk_adjustment_set_value( gtk_range_get_adjustment( GTK_RANGE( scale ) ), 
			      gtk_spin_button_get_value( GTK_SPIN_BUTTON( spin ) ) );
  }
}

void HSVPlugin::Scale2Spin( const gchar * scalename, const gchar * spinname ) const
{
  // get scale value and set it to spin
  GtkWidget * scale = glade_xml_get_widget( glade, scalename );
  GtkWidget * spin = glade_xml_get_widget( glade, spinname );
  if ( scale && spin ) {
    gtk_spin_button_set_value( GTK_SPIN_BUTTON( spin ),
			       gtk_adjustment_get_value( gtk_range_get_adjustment( GTK_RANGE( scale ) ) ) );
  }
}

// Export filter

extern "C" {

  GDKImageFilter *GetImageFilter( int index )
  {
    switch( index )
      {
      case 0: 
	return new HSVPlugin();
      }
    return NULL;
  }
  
}
